import asyncio
import aiohttp
from collections import Counter
from dataclasses import dataclass
import ipaddress
import logging
import platformdirs
import random
import re
import socket
import os
from sqlmodel import Session, SQLModel, create_engine, select
import warnings
from grep_ips.grep_ips import GrepIPs
from .parrots import Parrot


def cli(return_help_text=False):
    import argparse  # https://docs.python.org/3/library/argparse.html

    help_width = 78 if return_help_text else None  # consistent width for README.py
    formatter_class = lambda prog: argparse.HelpFormatter(
        prog,
        max_help_position=37,
        width=help_width,
    )
    parser = argparse.ArgumentParser(
        prog=app_name(),
        formatter_class=formatter_class,
    )
    parser.add_argument(
        "-t",
        "--timeout",
        type=int,
        default=12000,
        help="approximate timeout for http and https requests in milliseconds (default: 12000)",
    )
    parser.add_argument(
        "--minimum-match",
        type=int,
        default=2,
        help="an IP address is considered valid after this number of idential responses"
        + " (default: 2)",
    )
    parser.add_argument(
        "--overkill",
        type=int,
        default=0,
        help="number of initial requests to make beyond minimum-match (default: 0)",
    )
    parser.add_argument(
        "--majority-ratio",
        type=int,
        default=3,
        help="minimum ratio needed to overrule a conflicting response; must be an integer;"
        + " a value of 2 means 2:1, or that 6 responses of 8.7.8.7 are needed to overrule"
        + " 3 responses of 7.8.4.4 (default: 3)",
    )
    parser.add_argument(
        "--max-failures",
        type=int,
        default=10,
        help="maximum number of failed requests allowed (default: 10)",
    )
    parser.add_argument(
        "--max-connections",
        type=int,
        default=10,
        help="maximum number of simultaneous network connections allowed (default: 10)",
    )
    db_file_display = db_pathname().replace(os.path.expanduser('~'), '~')
    parser.add_argument(
        "--dbfile",
        type=str,
        default='',  # need to call db_pathname() again later with create_dir=True
        help=f"path for database file ('-' for memory-only; default: {db_file_display})",
    )
    parser.add_argument(
        "--show-api-providers",
        action='store_true',
        help="display the database of IP address API providers in a human-readable form and exit",
    )
    parser.add_argument(
        "-4",
        "--ipv4",
        action='store_true',
        help="use IPv4 only; note this or --ipv6 is highly recommended if both IPv4 and IPv6"
        + " are available, in order to avoid wasteful network traffic and unpredictable results"
        + " (sometimes --minimum-match IPv4 addresses will be received first, and"
        + " sometimes IPv6 will win)",
    )
    parser.add_argument(
        "-6",
        "--ipv6",
        action='store_true',
        help="use IPv6 only",
    )
    parser.add_argument(
        "-l",
        "--logfile",
        type=str,
        default='-',
        help="path for log file (default: write to STDERR)",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action='append_const',
        const=-1,
        dest="verbose",  # mapping:  "-q"->ERROR / ""->WARNING / "-v"->INFO / "-vv"->DEBUG
        help="silence warning messages",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action='append_const',
        const=1,
        help="increase verbosity",
    )
    if return_help_text:  # used by README.py
        return parser.format_help()
    args = parser.parse_args()
    if args.show_api_providers:
        engine = init_db(args.dbfile)
        Parrot.show_parrot_db(engine)
        return
    del args.show_api_providers
    if args.ipv4:
        args.ip_version = 4
    if args.ipv6:
        args.ip_version = 6
    del args.ipv4
    del args.ipv6
    args.log_level = 2 + (0 if args.verbose is None else sum(args.verbose))
    del args.verbose
    moa = MyoriginArgs(**args.__dict__)
    print(my_ip(moa))


@dataclass
class MyoriginArgs:  # see '--help' for descriptions
    timeout: int = 12000
    minimum_match: int = 2
    overkill: int = 0
    majority_ratio: int = 3
    max_failures: int = 10
    max_connections: int = 10
    dbfile: str = ''  # ''==db_pathname(), '-'==none, other==alternate file
    ip_version: int = 0  # 0==either, 4==IPv4 only, 6==IPv6 only
    logfile: str = '-'
    log_level: int = 0  # 0==disabled, 1==errors, 2==warnings, 3==info, 4==debug
    exception_level = 0  # 0==disabled, 1==raise exception for errors, 2==... for warnings, etc.


def app_name():
    return os.path.splitext(os.path.basename(__file__))[0]


class NetworkError(Exception):
    pass


def my_ip(args: MyoriginArgs) -> str:
    if args.log_level >= 3:  # info or debug
        log_format = '%(asctime)s.%(msecs)03d %(levelname)s %(message)s'
    else:
        log_format = '%(message)s'
    logging.basicConfig(
        format=log_format,
        datefmt='%H:%M:%S',
        filename=args.logfile if args.logfile != '-' else None,
        filemode='a',
    )
    logger = logging.getLogger('myorigin')
    log_levels = [logging.CRITICAL, logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]
    try:
        logger.setLevel(log_levels[args.log_level])
    except IndexError:
        logger.setLevel(logging.WARNING)
        error = "Invalid log level"
        logger.error(error)
        if args.exception_level >= 1:
            raise ValueError(error)
        return ""
    return asyncio.run(main_loop(args, logger))


def http_timeout(t):
    # ClientTimeout docs: https://docs.aiohttp.org/en/stable/client_reference.html#clienttimeout
    give_up_after = float(t) / 1000
    return aiohttp.ClientTimeout(
        # total=give_up_after,  # don't use this because it includes the time in queue
        sock_connect=give_up_after,
        sock_read=give_up_after,
    )


async def timer_mark(session, context, params):
    context.trace_request_ctx.append(asyncio.get_event_loop().time())


def http_timer():  # configure signals to record beginning and end of http requests
    trace_config = aiohttp.TraceConfig()
    # on_connection_queded_end → not called if a connection is immediately available
    # on_connection_reuseconn, on_connection_create_start → exactly one of these is always called
    # on_request_headers_sent.append → always called
    # on_request_end.append → always called
    trace_config.on_connection_reuseconn.append(timer_mark)  # start timer
    trace_config.on_connection_create_start.append(timer_mark)  # start timer
    trace_config.on_request_end.append(timer_mark)  # stop timer
    return trace_config


async def http_get(url: str, session: aiohttp.ClientSession, max_size=-1):
    if hasattr(http_get, 'pytest'):  # mock responses; see myorigin_test.py
        if http_get.pytest == 1:
            try:
                return re.match(r'http.*/(.*)', url).group(1), 290  # whatever follows final slash
            except:
                return f'test html response with no IP address', 180
        elif http_get.pytest == 2:
            raise ValueError("404 pytest simulated error")
    timer_marks = list()
    # change user agent string for better site compatibility and anonymity, though one site
    # ... suggests a Firefox user agent (https://www.whatismyip.com/automation-rules/);
    # Wget agent retrieved via: wget -O- -q http://httpbin.org/get |grep User-Agent
    headers = {'User-Agent': 'Wget/1.20.3 (linux-gnu)'}
    # headers = {'User-Agent': 'Python/3.8 aiohttp/3.8.1'}  # default on my system
    try:
        async with session.get(
            url,
            allow_redirects=False,
            trace_request_ctx=timer_marks,
            headers=headers,
        ) as response:
            if response.status != 200:
                raise aiohttp.ClientError(f"{response.status} {response.reason}")
            html = str((await response.content.read(max_size)).decode())
            assert len(timer_marks) == 2
            # SQLite max int is 2^63 ms, or 500000 requests per second for 100 years
            ms = round((timer_marks[1] - timer_marks[0]) * 1000)
            return html, ms
    # exceptions here seem to cause infinite loop (¿bug in my code?), e.g. trying "httpss://" url
    except Exception as e:  # most will be aiohttp.ClientError
        s = str(e)
        raise ValueError(s if len(s) > 1 else "Unknown exception in http_get()")


@dataclass
class FetchedIP:
    parrot: Parrot
    ip: str = ''  # IP address returned by server or error message
    ip_version: int = 0  # 0==none, 4==IPv4, 6==IPv6
    rtt: int = 0  # total milliseconds needed for request or 0 for error


async def get_ip(
    p: Parrot, session: aiohttp.ClientSession, q: asyncio.Queue, ip_version: int = 0
) -> None:
    fip = FetchedIP(p)
    a = None
    try:
        html, ms = await http_get(
            p.url(),
            session,
            max_size=16384,  # don't download it all
        )
    except ValueError as e:
        a = str(e)
        if a == '':
            a = "Unknown http_get() error"
    if a is None and len(html) < 3:
        a = f"Only {len(html)} bytes received"
    if a is None:
        ip = GrepIPs.grep_ips(
            html,
            global_ips_only=True,
            first_match_only=True,
            ip_version=ip_version,
        )
        if ip is not None:
            a = str(ip)
            fip.ip_version = 4 if type(ip) == ipaddress.IPv4Address else 6
            fip.rtt = ms
    if a is None:
        # call grep_ips() again to be able to give a more specific error message
        ip = GrepIPs.grep_ips(
            html,
            global_ips_only=False,
            first_match_only=True,
            ip_version=ip_version,
        )
        if ip is None:
            a = "No IP address found"
        else:
            a = f"Found non-global IP address {ip}"
    fip.ip = a
    await q.put(fip)


def weighted_random(options: dict):
    # return a randomly-chosen key with the probability represented by its value
    total = sum(options.values())
    for k, weight in options.items():
        if random.randint(0, total - 1) < weight:
            return k
        total -= weight
    assert False, f"bug in weighted_random, options {options}"


def are_we_there_yet(completed_jobs, args, logger):
    top_ip_count = 0  # max of IPv4 and IPv6
    top_ip_target = args.minimum_match
    for v in (4, 6):
        fail_count = sum(1 for r in completed_jobs if r.rtt == 0)
        if fail_count >= args.max_failures:
            raise NetworkError(f"{fail_count} requests failed; giving up")
        votes = Counter([r.ip for r in completed_jobs if (r.ip_version == v and r.rtt != 0)])
        if len(votes) == 0:
            continue
        ranked = sorted(votes.items(), key=lambda i: i[1], reverse=True)
        # ranked[0][0] is #1 candidate IP; ranked[0][1] is its vote count
        assert sum(1 for r in completed_jobs if (r.rtt != 0 and r.ip == '')) == 0
        if len(ranked) > 1 and not hasattr(are_we_there_yet, 'found_multiple_ips'):
            are_we_there_yet.found_multiple_ips = True  # only show once
            logger.info(f"multiple IPs received: {[k[0] for k in ranked]}")
        overrule = 0 if len(ranked) <= 1 else ranked[1][1] * args.majority_ratio
        target = max(args.minimum_match, overrule)  # votes necessary to win
        if ranked[0][1] >= target:  # #1 candidate has enough votes
            ip = ranked[0][0]
            msg = f"IP found: {ip} ({ranked[0][1]} successes,"
            logger.info(f"{msg} {fail_count} failures)")
            if len(ranked) > 1:  # retroactively mark other candidates as failed
                for r in completed_jobs:
                    if r.ip_version == v and r.rtt != 0 and r.ip != ip:
                        r.rtt = 0  # actually, it failed
                        r.ip = f'Incorrect IP returned: {r.ip}'
            raise DoneWithJobs(ip)
            # logger.info(f"{fail_count} fails for {len(completed_jobs)} jobs")
        top_ip_count = max(top_ip_count, ranked[0][1])
        top_ip_target = max(top_ip_target, target)
    if top_ip_count == 0:
        are_we_there_yet.__dict__.pop('found_multiple_ips', None)  # show again if needed
    return top_ip_count, top_ip_target


def log_debug_details(logger, top_ip_count, pending_jobs_count, top_ip_bonus, top_ip_target):
    if logger.isEnabledFor(logging.DEBUG):
        try:
            last_debug_msg = main_loop.last_debug_msg
        except AttributeError:
            last_debug_msg = ""
        msg = f"have {top_ip_count} IPs plus {pending_jobs_count} pending,"
        msg += f" want {top_ip_bonus}, need {top_ip_target}"
        if msg != last_debug_msg:  # avoid repeating the same message
            logger.debug(msg)
        main_loop.last_debug_msg = msg


class DoneWithJobs(Exception):
    pass


async def main_loop(args: MyoriginArgs, logger: logging.Logger) -> str:
    result = ""
    engine = init_db(args.dbfile)
    with Session(engine) as db_session:
        scores = dict()
        all_parrots = db_session.exec(select(Parrot))
        for p in all_parrots:
            if args.ip_version != 0 and args.ip_version != p.ip_version():  # only ones user wants
                continue
            score = p.score()
            if score <= 0:
                continue
            scores[p.id] = score
        q = asyncio.Queue()  # great tutorial: https://realpython.com/async-io-python/
        if args.ip_version == 0:
            ip_family = 0
        elif args.ip_version == 4:
            ip_family = socket.AF_INET
        elif args.ip_version == 6:
            ip_family = socket.AF_INET6
        else:
            assert False
        connector = aiohttp.TCPConnector(  # https://docs.aiohttp.org/en/stable/client_reference.html#tcpconnector
            limit=args.max_connections,
            family=ip_family,
        )
        completed_jobs = list()
        async with aiohttp.ClientSession(
            connector=connector, timeout=http_timeout(args.timeout), trace_configs=[http_timer()]
        ) as aio_session:
            try:
                logger.info(f"requests (need {args.minimum_match} matches):")
                pending_jobs_count = 0
                while True:  # spawn and collect jobs
                    # calculate what is still needed
                    top_ip_count, top_ip_target = are_we_there_yet(completed_jobs, args, logger)
                    top_ip_bonus = top_ip_target + args.overkill
                    jobs_count = top_ip_count + pending_jobs_count
                    # logging, overkill check
                    log_debug_details(
                        logger, top_ip_count, pending_jobs_count, top_ip_bonus, top_ip_target
                    )
                    if args.overkill > 0 and top_ip_bonus > jobs_count and len(scores) == 0:
                        msg = f"not enough providers for {top_ip_bonus} requests"
                        logger.warning(f"{msg}; ignoring '--overkill'")
                        args.overkill = 0
                        top_ip_bonus = top_ip_target + args.overkill
                    # spawn another get_ip() job if needed
                    if top_ip_bonus > jobs_count and pending_jobs_count < args.max_connections:
                        # TCPConnector(limit=...) does this too, but checking args.max_connections
                        # ... here delays "not enough providers" so we can collect more responses
                        if len(scores) == 0:
                            error = f"not enough providers for {args.minimum_match} matches"
                            raise NetworkError(error)
                        p_id = weighted_random(scores)
                        statement = select(Parrot).where(Parrot.id == p_id)
                        p = db_session.exec(statement).one_or_none()
                        asyncio.create_task(get_ip(p, aio_session, q, ip_version=args.ip_version))
                        pending_jobs_count += 1
                        logger.debug(f"launching task {p.id}")
                        p.del_keys_of_same_parrot(scores, logger)  # don't use this parrot again
                        main_loop.last_debug_msg = ""  # repeat debug msg after launch even if same
                    await asyncio.sleep(0.0001)
                    # process completed jobs
                    try:
                        r: FetchedIP = q.get_nowait()  # may raise QueueEmpty
                        completed_jobs.append(r)  # keep a list to write to DB later
                        if r.rtt != 0:  # got valid IP
                            # note success count will be wrong if IP is later found to be wrong
                            p = f"{r.parrot.success_count + 1} of {r.parrot.attempt_count + 1}"
                            msg = f"    {r.parrot.url()} → {r.ip} ({r.rtt} ms;"
                            logger.info(f"{msg} {p} succeeded)")
                        else:
                            p = f"{r.parrot.success_count} of {r.parrot.attempt_count + 1}"
                            msg = f"    {r.parrot.url()} → {r.ip[:40]}"
                            logger.info(f"{msg} ({p} succeeded)")
                        pending_jobs_count -= 1
                    except asyncio.QueueEmpty:
                        await asyncio.sleep(0.005)  # reduce looping when waiting on responses
                    await asyncio.sleep(0.0001)
            except DoneWithJobs as e:
                result = str(e)
            except NetworkError as e:
                logger.error(e)
                if args.exception_level >= 1:
                    raise NetworkError(e)
        await connector.close()
        for r in completed_jobs:
            r.parrot.attempt_count += 1
            if r.rtt != 0:  # got valid IP of some sort
                r.parrot.success_count += 1
                r.parrot.total_rtt += r.rtt
            else:
                r.parrot.last_errmsg = r.ip
            db_session.add(r.parrot)
            db_session.commit()
    return result


def mkdir_r(path):  # like Linux `mkdir --parents`
    if path == '':
        return
    base = os.path.dirname(path)
    if not os.path.exists(base):
        mkdir_r(base)
    os.makedirs(path, exist_ok=True)
    # except (PermissionError, FileNotFoundError, NotADirectoryError):
    #     one of these will be raised if the directory cannot be created


def db_pathname(create_dir=False):
    config_dir = platformdirs.user_config_dir(app_name())
    if create_dir:
        mkdir_r(config_dir)
    return os.path.join(config_dir, f'data.sqlite')


def init_db(db_file):
    if hasattr(init_db, 'engine'):  # only need to initialize once
        return init_db.engine
    warnings.filterwarnings(  # https://github.com/tiangolo/sqlmodel/issues/189#issuecomment-1018014753
        "ignore", ".*Class SelectOfScalar will not make use of SQL compilation caching.*"
    )
    if db_file == '':  # use default location
        db_file = db_pathname(create_dir=True)
    elif db_file == '-':
        db_file = ':memory:'
    init_db.engine = create_engine(f'sqlite:///{db_file}', echo=False)
    SQLModel.metadata.create_all(init_db.engine)
    Parrot.startup(init_db.engine)
    return init_db.engine


if __name__ == "__main__":
    cli()



from dataclasses import dataclass
from enum import Enum
import logging
import re
from sqlmodel import Field, Session, SQLModel, create_engine, select


class Parrot(SQLModel, table=True):  # data for one interface of an API provider
    # (Parrots repeats what they hear. A my-IP API provider repeats the caller's IP back to them.)
    id: str = Field(primary_key=True)  # e.g. "0b.ipapi.co"; {ip_version}{proto}.{parrot_name}
    address: str = ""  # host+path (URL without protocol)
    milliweight: int = 1000  # 2000 means 2x more likely to be used; 0 means disabled
    attempt_count: int = 0  # number of attempted connections
    success_count: int = 0  # number of valid IP addresses returned
    total_rtt: int = 0  # total rtt (in ms) for all successful attempts
    last_errmsg: str = ""

    def ip_version(self) -> int:  # always 4 or 6
        return int(self.id[0])

    def proto(self) -> str:  # always 'p' (http) or 's' (https)
        return self.id[1]

    def url(self):
        if self.proto() == 'p':
            return 'http://' + self.address
        if self.proto() == 's':
            return 'https://' + self.address
        assert ValueError, f"Not yet implemented"

    @dataclass
    class ParrotFields:
        name: int  # unique name so we can safely copy future changes to user's DB
        address: str  # URL without 'https://' or 'http://'
        ip_version: int  # 0==both, 4==IPv4 only, 6==IPv6 only
        proto: str  # p==http only, s==https only, b==both, x==disabled

    @staticmethod
    def parrot_data(flock_data: str) -> ParrotFields:
        for line in flock_data.split('\n'):
            wout_comments = re.sub(r'( +|^)#.*\n?', '', line)  # strip comments
            if len(wout_comments) == 0:
                continue
            fields = re.sub(r' +', ' ', wout_comments).split(' ')
            assert len(fields) == 4, f"invalid flock_data line: {line}"
            assert fields[2] in ('0', '4', '6'), f"invalid flock_data line: {line}"
            assert fields[3] in ('p', 's', 'b', 'x'), f"invalid flock_data line: {line}"
            yield Parrot.ParrotFields(
                name=fields[0],
                address=fields[1],
                ip_version=int(fields[2]),
                proto=fields[3],
            )

    @staticmethod
    def startup(engine):
        with Session(engine) as session:
            statement = select(Parrot).where(Parrot.id == '4p.__version__')
            record = session.exec(statement).one_or_none()
            flock_data_version = 'v10094'  # add 1 when a change is made to flock_data.py
            if record and record.address >= flock_data_version:  # database is up-to-date
                return
            logger = logging.getLogger('myorigin')
            if record is not None:
                msg = "updating flock_data_version"
                logger.debug(f"{msg} from {record.address} to {flock_data_version}")
            else:
                logger.debug(f"updating flock_data_version from (none) to {flock_data_version}")
            if record is None:
                record = Parrot()
                record.id = '4p.__version__'
                record.milliweight = 0  # ensure we don't try to connect
            record.address = flock_data_version
            session.add(record)
            if hasattr(Parrot.startup, 'pytest'):  # mock responses; see myorigin_test.py
                flock_data = Parrot.startup.pytest
            else:
                from .flock_data import flock_data
            for row in Parrot.parrot_data(flock_data):  # one row in flock_data text
                for v in (4, 6):  # IPv4, IPv6
                    for p in ('p', 's'):  # http, https
                        id = f'{v}{p}.{row.name}'
                        proto_active = row.proto == p or row.proto == 'b'
                        ip_version_active = row.ip_version == 0 or row.ip_version == v
                        statement = select(Parrot).where(Parrot.id == id)
                        record = session.exec(statement).one_or_none()
                        if record is None:  # no existing record in DB
                            if not (proto_active and ip_version_active):
                                continue  # no need to add disabled records
                            record = Parrot()  # add new record to database
                            record.id = id
                            logger.debug(f"adding record {id}")
                        if proto_active and ip_version_active:
                            record.milliweight = 1000  # enable in case it was previously disabled
                        else:
                            record.milliweight = 0  # disable in DB
                        if record.address != row.address and record.address != "":
                            msg = f"updating record {id}"
                            logger.debug(f"{msg} from {record.address} to {row.address}")
                        record.address = row.address
                        session.add(record)
            session.commit()

    @staticmethod
    def acive_parrot_count():
        total = 0
        if True:
            from .flock_data import flock_data
        for p in Parrot.parrot_data(flock_data):
            if p.proto in ('p', 's', 'b'):
                total += 1
        return total

    def score(self):  # compute a score which will determine how likely it is to be chosen
        if self.attempt_count != 0:
            percent = 100.0 * self.success_count / self.attempt_count
        else:
            percent = 100.0
        score = round(percent**2 / 100.0)  # biggest portion of score is success rate
        score += 5  # every parrot gets a small chance of being selected
        if self.attempt_count < 10:
            score += 5  # prefer new parrots
        if self.success_count != 0:
            average_ms = 1.0 * self.total_rtt / self.success_count
            score += round(max((5000 - average_ms) / 400, 0.0))  # prefer faster parrots
        score *= self.milliweight  # normally 1000, but can be 0 to disable or more to promote
        return score

    def del_keys_of_same_parrot(self, scores_dict, logger=None):
        p_name = self.id[3:]  # parrot name without ip_version, proto
        ids_to_del = [k for k in scores_dict.keys() if k[3:] == p_name]
        if logger:
            logger.debug(f"removing {', '.join(ids_to_del)}")
        for k in ids_to_del:  # eliminate all IDs from same flock_data line, e.g.
            del scores_dict[k]  # don't check both http://ident.me and https://ident.me

    @staticmethod
    def show_parrot_db(engine):
        with Session(engine) as session:
            to_show = dict()
            results = session.exec(select(Parrot))
            for record in results:
                score = record.score()
                if score <= 0:
                    continue  # don't display disabled records
                disp = ""
                disp += f"ipv{record.ip_version()}.{record.url()}:\n"
                if record.attempt_count != 0:
                    percent = round(100.0 * record.success_count / record.attempt_count)
                    disp += f"    {percent}% success rate"
                    disp += f" ({record.success_count} of {record.attempt_count})\n"
                else:
                    disp += f"    not yet attempted\n"
                if record.success_count != 0:
                    average_ms = round(1.0 * record.total_rtt / record.success_count)
                    disp += f"    {average_ms} ms average round trip time\n"
                if len(record.last_errmsg) > 0:
                    disp += f"    most recent error: {record.last_errmsg[:70]}\n"
                disp += f"    score: {round(score/1000):,} points\n"
                to_show[f'{score:08}.{record.id}'] = disp  # sortable, unique key
            for key in dict(sorted(to_show.items(), reverse=True)):
                print(to_show[key], end="")



flock_data = '''
#field 0: unique parrot name
#field 1: address without http or https  
#field 2: ip_version (0 (both) or 4 or 6)
#field 3: protocol (p (http only) or s (https only) or b (both) or x (disabled))
#everything after # is a comment

# when making changes, be sure to increment flock_data_version
2ip.me                api.2ip.me/provider.json?ip=     0 x # 429 Too Many Requests
2ip.ru                2ip.ru/                          0 x # No IP address found
2ip.tools             2ip.tools/                       0 x # Cannot connect to host 2ip.tools:80 ...
2ip.ua                api.2ip.ua/geo.json?ip=          0 x # 429 Too Many Requests
42.pl                 ip.42.pl/                        4 b
66.171.248.178        66.171.248.178/                  0 x # Cannot connect to host 66.171.248.1 ...
7fw.de                7fw.de/ipraw.php                 4 b
89.39.105.12          89.39.105.12/ip.shtml            0 x # No IP address found
aboutmyip.com         aboutmyip.com/                   0 x # No IP address found
amazonaws.com         checkip.amazonaws.com/           4 b
am.i.mullvad.net      am.i.mullvad.net/ip              0 s
amulex.com            ip.amulex.com/                   0 x # No IP address found
anysrc.net            ip.anysrc.net/                   0 b
appspot.com           ip.appspot.com/                  0 x # 503 Service Unavailable
aruljohn.com          ip.aruljohn.com/json             0 s
avast.com             ip-info.ff.avast.com/            0 x # 404 Not Found
aws.a2z.com           checkip.check-ip.aws.a2z.com/    4 b
bahoot.com            bahoot.com/                      0 x # Connection timeout to host http://b ...
bearsmyip.com         disabled.bearsmyip.com/          0 x # Cannot connect to host disabled.bea ...
bigdatacloud.net      api.bigdatacloud.net/data/client-ip 4 b
cctv.pk               ip.cctv.pk/                      0 x # No IP address found
checkip.me            checkip.me/                      0 x # Cannot connect to host checkip.me:8 ...
checkip.net           checkip.net/                     4 s
checkip.org           checkip.org/                     4 b
checkip.pw            checkip.pw/                      0 x # Cannot connect to host checkip.pw:8 ...
checkmyip.com         checkmyip.com/                   0 x # Timeout on reading data from socket
check-my-ip.net       check-my-ip.net/                 0 x # Connection timeout to host http://c ...
chinaz.com            ip.chinaz.com/                   0 x # [Errno 104] Connection reset by peer
cloudflare.com        www.cloudflare.com/cdn-cgi/trace 0 b
cmyip.com             cmyip.com/                       0 x # Cannot connect to host cmyip.com:44 ...
cmyip.net             cmyip.net/                       0 x # Cannot connect to host cmyip.net:80 ...
codeluxsoftware.com   codeluxsoftware.com/             0 x # Cannot connect to host codeluxsoftw ...
crymyip.com           crymyip.com/                     4 s
curlmyip.com          curlmyip.com/                    0 x # Connection timeout to host http://c ...
curlmyip.net          curlmyip.net/                    0 b
dawhois.com           dawhois.com/my_ip_address.php    0 s
db-ip.com             api.db-ip.com/                   0 x # 302 Found
dnsdynamic.org        myip.dnsdynamic.org/             0 x # Cannot connect to host myip.dnsdyna ...
dns.he.net            checkip.dns.he.net/              0 b
dnsomatic.com         myip.dnsomatic.com/              4 b
dnswatch.info         www.dnswatch.info/what's-my-ip   0 x # No IP address found
dpool.sina.com.cn     dpool.sina.com.cn/               0 x # Connection timeout to host http://d ...
dramor.net            myip.dramor.net/                 0 x # 404 no such forward entry
dtdns.com             myip.dtdns.com/                  0 x # Cannot connect to host myip.dtdns.c ...
dyn.com               checkip.dyn.com/                 0 x # 502 Bad Gateway
dyndns.com            checkip.dyndns.com/              0 x # 502 Bad Gateway
dyndns.es             checkip.dyndns.es/               0 x # 403 Forbidden
dyndns.it             checkip.dyndns.it/               4 b
dyndns.org            checkip.dyndns.org/              4 p
dynupdate.no-ip.com   ip1.dynupdate.no-ip.com/         4 p
easylife.tw           myip.easylife.tw/                0 x # Connection timeout to host http://m ...
e-localizaip.com      e-localizaip.com/                0 x # Cannot connect to host e-localizaip ...
eth0.me               eth0.me/                         4 b
everdot.org           whatismyip.everdot.org/          0 x # Cannot connect to host whatismyip.e ...
extreme-ip-lookup.com extreme-ip-lookup.com/json/?key=demo2 4 b
findmyipaddress.com   findmyipaddress.com/             0 x # Connection timeout to host https:// ...
findmyip.org          findmyip.org/                    0 x # Timeout on reading data from socket
formyip.com           formyip.com/                     0 x # 403 Forbidden
freegeoip.app         freegeoip.app/                   0 x # No IP address found
freegeoip.live        freegeoip.live/xml/              0 s
game.qq.com           apps.game.qq.com/comm-htdocs/ip/get_ip.php 4 b
geodatatool.com       www.geodatatool.com/             4 s
geoip.co.uk           disabled.geoip.co.uk/            0 x # Cannot connect to host disabled.geo ...
geoip-db.com          geoip-db.com/                    0 x # Cannot connect to host geoip-db.com ...
geoiptool.com         geoiptool.com/                   0 x # Cannot connect to host geoiptool.co ...
geoip.vmn.net         geoip.vmn.net/                   0 x # No IP address found
geolocation-db.com    geolocation-db.com/json/         4 b
geoplugin.net         www.geoplugin.net/csv.gp         4 p
get.geojs.io          get.geojs.io/                    0 s
get-myip.com          get-myip.com/                    4 s
getmyip.co.uk         getmyip.co.uk/                   0 x # 302 Moved Temporarily
getmyip.org           getmyip.org/                     0 x # No IP address found
getmyip.win           getmyip.win/                     0 x # Cannot connect to host getmyip.win: ...
google.com            www.google.com/search?q=my+ip    0 x # 403 Forbidden
heltech.se            myip.heltech.se/                 4 b
herokussl.com         nagano-19599.herokussl.com       0 x # Cannot connect to host nagano-19599 ...
hostip.info           api.hostip.info/get_html.php     4 b
httpbin.org           httpbin.org/ip                   4 b
icanhazip.com         icanhazip.com/                   0 b
ident.me              ident.me/                        0 b # https://api.ident.me/
ifcfg.me              ifcfg.me/                        0 x # Cannot connect to host ifcfg.me:443 ...
ifconfig.co           ifconfig.co/                     0 b
ifconfig.io           ifconfig.io/ip                   0 b
ifconfig.me           ifconfig.me/ip                   4 b
ilmioip.it            ilmioip.it/                      0 x # 302 Found
indirizzo-ip.com      www.indirizzo-ip.com/IP-Whois.php 4 b
inet-ip.info          inet-ip.info/                    4 b
infoip.io             api.infoip.io/                   4 b # https://ciokan.docs.apiary.io/
ip138.com             disabled.ip138.com/              0 x # Cannot connect to host disabled.ip1 ...
ip-1.com              ip-1.com/                        0 x # Cannot connect to host ip-1.com:443 ...
ip2location.com       www.ip2location.com/             0 x # No IP address found
ip2nation.com         ip2nation.com/                   4 p
ip4.me                ip4.me/                          4 b
ip6.me                ip6.me/api/                      0 b
ip-addr.es            ip-addr.es/                      4 s
ip-address.cc         ip-address.cc/                   0 x # 302 Found
ipaddresscheck.com    www.ipaddresscheck.com/          0 x # 'utf-8' codec can't decode byte 0x9 ...
ipaddress.com         www.ipaddress.com/               0 x # API discontinued?
ipaddress.my          www.ipaddress.my/                0 s
ipaddress.net         ipaddress.net/                   0 x # No IP address found
ipaddress.org         disabled.ipaddress.org/          0 x # Cannot connect to host disabled.ipa ...
ip-address.ru         ip-address.ru/                   0 x # No IP address found
ipaddress.sh          ipaddress.sh/                    4 b
ipaddressworld.com    ipaddressworld.com/              0 x # 403 Forbidden
ip-adress.com         www.ip-adress.com/what-is-my-ip-address 0 x # No IP address found
ip-adress.eu          www.ip-adress.eu/                0 b
ipapi.co              ipapi.co/ip                      0 s
ip-api.com            ip-api.com/line/?fields=query    4 p
ipchecker.info        ipchecker.info/                  0 x # 302 Found
ip-check.info         ip-check.info/                   0 x # Cannot connect to host ip-check.inf ...
ipchicken.com         www.ipchicken.com/               4 s
ip.cn                 www.ip.cn/                       0 x # No IP address found
ipcode.pw             ipcode.pw/                       0 x # Only 0 bytes received
ip-detect.net         ip-detect.net/                   0 x # Cannot connect to host ip-detect.ne ...
ipecho.net            ipecho.net/plain                 4 b
ipify.org             api.ipify.org/                   4 b
ipinfodb.com          www.ipinfodb.com/                0 b
ipinfo.info           api.ipapi.com/api/?access_key=   0 x # No IP address found
ipinfo.io             ipinfo.io/ip                     4 b
ip-info.org           ip-info.org/                     0 x # 307 Temporary Redirect
ip-info.xyz           ip-info.xyz/                     0 x # Timeout on reading data from socket
ipip.net              myip.ipip.net/                   4 b
ipleak.net            ipleak.net/                      0 b
iplocation.net        iplocation.net/                  0 x # No IP address found
iplogger.org          iplogger.org/myip/               0 x # No IP address found
iplogger.ru           iplogger.ru/us/myip/             0 x # No IP address found
ipmonkey.com          ipmonkey.com/                    4 p
ip.nf                 ip.nf/me.txt                     4 s
ipof.in               ipof.in/txt                      0 s
ipogre.com            ipv4.ipogre.com/                 0 x # Connection timeout
ip-ping.ru            www.ip-ping.ru/                  0 x # 'utf-8' codec can't decode byte 0xd ...
ipregistry.co         api.ipregistry.co/?key=tryout    0 b # https://ipregistry.co/docs/
ip.sb                 api.ip.sb/ip                     0 b # https://ip.sb/api/
ip-score.com          ip-score.com/ip                  4 s # https://ip-score.com/api
ip-secrets.com        www.ip-secrets.com/              0 x # 'utf-8' codec can't decode byte 0xa ...
ipstack.com           ipstack.com/                     0 s
iptrackeronline.com   www.iptrackeronline.com/         0 s
ipv6-test.com         ipv6-test.com/api/myip.php       0 b
ip-who-is.com         ip-who-is.com/                   0 x # Cannot connect to host ip-who-is.co ...
ip-whois.net          ip-whois.net/                    0 x # 'utf-8' codec can't decode byte 0xc ...
israel.net            myip.israel.net/                 0 x # 302 Found
ivpn.net              www.ivpn.net/                    0 s
jacware.com           myip.jacware.com/                0 x # Cannot connect to host myip.jacware ...
j.maxmind.com         j.maxmind.com/                   0 x # Cannot connect to host j.maxmind.co ...
jsonip.com            jsonip.com/                      0 s
keliweb.it            keliweb.it/mioip.php             0 x # 403 Forbidden
l2.io                 l2.io/ip.json                    4 b
lawrencegoetz.com     www.lawrencegoetz.com/programs/ipinfo/ 4 p
localizaip.com.br     localizaip.com.br/               0 x # Cannot connect to host localizaip.c ...
meip.eu               meip.eu/                         4 s
meuip.net.br          meuip.net.br/                    4 s
mioip.biz             mioip.biz/                       4 s
mioip.ch              mioip.ch/                        4 p
mioip.info            mioip.info/                      4 b
mio-ip.it             www.mio-ip.it/                   0 x # No IP address found
mioip.it              mioip.it/                        4 b
mioip.org             mioip.org/                       0 x # Cannot connect to host mioip.org:80 ...
mioip.win             mioip.win/                       0 x # 526 
moanmyip.com          www.moanmyip.com/                0 s
mon-ip.com            www.mon-ip.com/info-adresse-ip.php 0 x # 403 Forbidden
mudfish.net           myip.mudfish.net/                4 b
mycamip.com           www.mycamip.com/                 0 s
myexternalip.com      myexternalip.com/raw             4 b
myglobalip.com        myglobalip.com/                  0 x # Cannot connect to host myglobalip.c ...
myipaddress.com       www.myipaddress.com/             0 b
my-ip-address.co      www.my-ip-address.co/            0 s
my-ip-address.net     disabled.my-ip-address.net       0 x # Cannot connect to host disabled.my- ...
myip.am               myip.am/                         0 x # No IP address found
myip.by               myip.by/                         0 x # Cannot connect to host myip.by:80 s ...
myip.cc               disabled.myip.cc/                0 x # Cannot connect to host disabled.myi ...
myip.cf               myip.cf/                         0 s
myip.ch               myip.ch/                         0 x # Cannot connect to host myip.ch:443  ...
my-ip.club            my-ip.club/                      0 x # Cannot connect to host my-ip.club:4 ...
myip.cn               myip.cn/                         0 x # Cannot connect to host myip.cn:80 s ...
myip.co.il            myip.co.il/                      4 p
myip.com              api.myip.com/                    0 s
myip.com.br           myip.com.br/                     0 x # No IP address found
myip.com.tw           myip.com.tw/                     4 s
myip.com.ua           myip.com.ua/                     0 x # Found non-global IP address 127.0.0.1
myip.co.nz            myip.co.nz/                      0 x # Cannot connect to host myip.co.nz:8 ...
myip.cz               myip.cz/                         4 s
myip.dk               disabled.myip.dk/                0 x # Cannot connect to host disabled.myi ...
myip.es               myip.es/mi-ip                    0 x # 403 Forbidden
myip.eu               myip.eu/                         0 x # No IP address found
myip.fi               myip.fi/                         4 s
myip.gratis           myip.gratis/                     0 x # Cannot connect to host myip.gratis: ...
myip.gr               disabled.myip.gr/                0 x # Cannot connect to host disabled.myi ...
myip.ht               myip.ht/                         0 b
myip.info             ipv6.myip.info/                  6 s
myipinfo.net          myipinfo.net/                    0 x # 403 Forbidden
my-ip.io              api.my-ip.io/ip                  0 s # https://www.my-ip.io/api
myip.io               disabled.myip.io/                0 x # Cannot connect to host disabled.myi ...
myip.is               myip.is/                         0 x # 'utf-8' codec can't decode byte 0xd ...
myipis.net            myipis.net/                      0 x # No IP address found
myip.knet.ca          myip.knet.ca/                    4 s
myip.kz               myip.kz/                         4 s
myip.ma               myip.ma/                         4 b
myip.ms               api.myip.ms/ip                   0 x # No IP address found
myip.mx               myip.mx/                         0 x # Connection timeout to host https:// ...
myip.net              myip.net/                        0 x # Connection timeout to host http://m ...
myip.nl               myip.nl/                         4 b
myipnow.com           myipnow.com/                     4 s
myipnumber.com        myipnumber.com/                  4 p
myip.nu               myip.nu/                         0 x # No IP address found
myiponline.com        myiponline.com/                  0 x # Connection timeout to host http://m ...
myip.report           myip.report/                     0 x # No IP address found
myip.ru               myip.ru/                         0 x # No IP address found
myip.se               myip.se/                         0 x # 403 Forbidden
myip.si               myip.si/                         0 x # No IP address found
myip.tk               myip.tk/                         0 x # Cannot connect to host myip.tk:80 s ...
myip.tw               myip.tw/                         0 x # No IP address found
myip.wtf              text.myip.wtf/                   0 b
myip.zone             disabled.myip.zone/              0 x # Cannot connect to host disabled.myi ...
mylnikov.org          disabled.mylnikov.org/           0 x # Connection timeout to host http://d ...
mylocation.org        mylocation.org/                  4 s
mysau.com.au          myip.mysau.com.au/               4 b
my-proxy.com          www.find-ip.net/                 4 b
mypubip.com           mypubip.com/                     0 b
narak.com             checkip.narak.com/               0 x # Cannot connect to host checkip.nara ...
network-tools.com     network-tools.com/               0 x # No IP address found
nmonitoring.com       myip.nmonitoring.com/            0 x # No IP address found
northstate.net        myip.northstate.net/             4 p
opendns.com           myip.opendns.com/                0 x # Cannot connect to host myip.opendns ...
ozymo.com             myip.ozymo.com/                  0 x # No IP address found
qualmeuip.com.br      qualmeuip.com.br/meuip/          0 x # Timeout on reading data from socket
readip.info           readip.info/                     0 x # No IP address found
rest7.com             api.rest7.com/v1/my_ip.php       4 p
rs.sr                 myip.rs.sr/                      4 s
sdu.dk                myip.sdu.dk/                     4 s
searchenginereports.net searchenginereports.net/what-is-my-ip 0 x # No IP address found
seeip.org             ip.seeip.org/                    0 s
sfml-dev.org          www.sfml-dev.org/ip-provider.php 4 p
shmyip.com            shmyip.com/                      0 x # Connection timeout to host https:// ...
shorty.org            myip.shorty.org/                 0 x # Cannot connect to host myip.shorty. ...
show-ip.com           show-ip.com/                     0 x # No IP address found
showipinfo.net        showipinfo.net/                  0 x # Cannot connect to host showipinfo.n ...
showip.net            showip.net/                      4 b
showmemyip.com        www.showmemyip.com/              4 s
showmyipaddress.com   www.showmyipaddress.com/         0 x # No IP address found
showmyipaddress.eu    wwww.showmyipaddress.eu/         0 x # Cannot connect to host wwww.showmyi ...
showmyip.com.ar       www.showmyip.com.ar/             0 p
showmyip.com          www.showmyip.com/                4 s
showmyip.co.uk        disabled.showmyip.co.uk/         0 x # Cannot connect to host disabled.sho ...
show-my-ip.de         www.show-my-ip.de/ipadresse/     0 x # Timeout on reading data from socket
showmyip.gr           www.showmyip.gr/                 0 s
showmyipnow.com       showmyipnow.com/                 0 x # Cannot connect to host showmyipnow. ...
shtuff.it             shtuff.it/myip/short/            0 x # Cannot connect to host shtuff.it:80 ...
smart-ip.net          smart-ip.net/myip                0 x # Connection timeout to host http://s ...
surfeasy.com          myip.surfeasy.com/               0 x # Cannot connect to host myip.surfeas ...
sypexgeo.net          api.sypexgeo.net/                4 b
taobao.com            ip.taobao.com/                   0 x # No IP address found
telespex.com          myip.telespex.com/               0 x # 302 Moved Temporarily
telize.com            ip.telize.com/                   0 x # Cannot connect to host ip.telize.co ...
tell-my-ip.com        disabled.tell-my-ip.com/         0 x # No IP address found
test-ipv6.com         test-ipv6.com/ip/                4 b # ironically, no IPv6 AAAA record
thesafety.us          thesafety.us/check-ip            0 x # No IP address found
tnedi.me              tnedi.me/                        0 b # https://ipa.tnedi.me/
tnx.nl                tnx.nl/ip                        0 b
tool.la               ip.tool.la/                      0 x # Connection timeout to host http://i ...
toolsyep.com          toolsyep.com/en/what-is-my-ip/   0 x # No IP address found
tracemyip.com         tracemyip.com/                   4 s
tracemyip.org         www.tracemyip.org/               0 x # No IP address found
trackip.net           www.trackip.net/ip               0 b
tunnelbear.com        www.tunnelbear.com/whats-my-ip   0 x # No IP address found
tyk.nu                ip.tyk.nu/                       0 b
ua.edu                myip.ua.edu/                     0 x # Connection timeout to host http://m ...
uconn.edu             myip.uconn.edu/                  0 x # Cannot connect to host myip.uconn.e ...
ultratools.com        ultratools.com/                  0 x # 302 Found
urls.is               ip.urls.is/                      4 p
utrace.de             utrace.de/                       0 x # Connection timeout to host https:// ...
v6shell.org           myip.v6shell.org/                0 x # Cannot connect to host myip.v6shell ...
vermiip.es            vermiip.es/                      4 b
viewmyip.eu           viewmyip.eu/                     4 s
vinflag.com           vinflag.com/                     0 x # No IP address found
webmasterhome.cn      ip.webmasterhome.cn/             0 x # Connection timeout to host http://i ...
websupport.sk         ip.websupport.sk/                0 b
wgetip.com            wgetip.com/                      0 b
whatismybrowser.com   www.whatismybrowser.com/         0 x # No IP address found
whatismyipaddress.com bot.whatismyipaddress.com/       0 x # discontinued "due to massive abuse"
whatismyip.akamai.com whatismyip.akamai.com/           4 p
whatismyip.ca         whatismyip.ca/                   0 x # Cannot connect to host whatismyip.c ...
whatismyip.com        automation.whatismyip.com/n09230945.asp 0 x # API discontinued?
whatismyip.com.br     whatismyip.com.br/               0 x # 403 Forbidden
whatismyip.li         disabled.whatismyip.li/home      0 x # Cannot connect to host disabled.wha ...
whatismyip.net        www.whatismyip.net/              0 x # No IP address found
whatismyip.org        www.whatismyip.org/my-ip-address 0 s
whatismypublicip.com  www.whatismypublicip.com/        4 s
whatmyip.us           whatmyip.us/                     0 x # Cannot connect to host whatmyip.us: ...
whatsmyipaddress.com  whatsmyipaddress.com/            0 x # No IP address found
whatsmyipaddress.net  whatsmyipaddress.net/            0 x # No IP address found
whats-my-ip-address.org whats-my-ip-address.org/       0 x # 302 Found
w.hatsmyip.com        w.hatsmyip.com/                  0 x # Cannot connect to host w.hatsmyip.c ...
whatsmyip.ie          whatsmyip.ie/                    4 b
whatsmyip.net         whatsmyip.net/                   0 b
whatsmyip.org         www.whatsmyip.org/               0 x # Connection timeout to host http://w ...
whatsmyip.us          whatsmyip.us/                    0 x # Cannot connect to host whatsmyip.us ...
whatsmyip.website     whatsmyip.website/               0 x # 302 Found
whereisip.net         disabled.whereisip.net/          0 x # No IP address found
whoer.me              whoer.me/                        0 x # No IP address found
whoer.net             whoer.net/                       4 s
whoisping.com         whoisping.com/                   0 x # No IP address found
wipmania.com          api.wipmania.com/                0 x # Cannot connect to host api.wipmania ...
wtfismyip.com         wtfismyip.com/text               0 b
xmyip.com             www.xmyip.com/                   4 s
yougetsignal.com      www.yougetsignal.com/what-is-my-ip-address/ 4 s
youip.net             youip.net/                       4 s
your-ip-address.com   your-ip-address.com/             4 b
your-ip-fast.com      your-ip-fast.com/                0 x # Cannot connect to host your-ip-fast ...
yourip.us             www.yourip.us/                   0 x # 302 Found
zx2c4.com             zx2c4.com/ip                     0 b
'''



import ipaddress
import re


class GrepIPs:
    def grep_ips(text, ip_version=0, global_ips_only=False, first_match_only=False):
        """
        If first_match_only == True, returns ipaddress.IPv4Address or ipaddress.IPv6Address or None
        If first_match_only == False, returns a possibly empty list of IPv4Address or IPv6Address
        """
        # see also https://gist.github.com/dfee/6ed3a4b05cfe7a6faf40a2102408d5d8
        ipv4 = r'(?<![0-9a-zA-Z.])(?:[0-9]{1,3}\.){3}[0-9]{1,3}(?![0-9a-zA-Z.])'
        ipv6 = r'(?<![0-9a-zA-Z:])(?:[0-9a-f]{0,4}:){1,7}[0-9a-f]{0,4}(?![0-9a-zA-Z:])'
        if ip_version == 4:
            ip_re = re.compile(f'({ipv4})')
        elif ip_version == 6:
            ip_re = re.compile(f'({ipv6})')
        else:
            ip_re = re.compile(f'((?:{ipv4})|(?:{ipv6}))')
        if not first_match_only:
            result = list()
        for m in ip_re.finditer(text):
            try:
                i = ipaddress.ip_address(m.group(1))
            except ValueError:  # e.g. 333.4.5.6
                continue
            if global_ips_only and (not i.is_global):  # e.g. 192.168.1.100
                continue
            if first_match_only:
                return i
            result.append(i)
        if first_match_only:
            return None
        else:
            return result



import random
import textwrap
import myorigin.myorigin


def test_my_ip_set1(caplog) -> None:
    myorigin.myorigin.init_db.__dict__.pop('engine', None)  # force Parrot.startup() to re-run
    test_flock_data = '''
        site91.com     site91.com/4.5.6.7           4 b
        site62.com     site62.com/xyxyxyxyxy        4 b
        site98.com     site98.com/xyxyxyxyxy        4 b
        site97.com     site97.com/xyxyxyxyxy        4 b
        site94.com     site94.com/xyxyxyxyxy        4 b
        site93.com     site93.com/xyxyxyxyxy        4 b
        site80.com     site80.com/4.5.6.7           4 b
        site29.com     site29.com/xyxyxyxyxy        4 b
        site55.com     site55.com/4.5.6.7           4 b
        site44.com     site44.com/xyxyxyxyxy        4 b
        site85.com     site85.com/ip_is_4.5.6.7_    4 b
        site43.com     site43.com/2604::89          6 b
    '''
    myorigin.parrots.Parrot.startup.pytest = textwrap.dedent(test_flock_data)
    myorigin.myorigin.http_get.pytest = 1  # mock http_get()
    args = myorigin.MyoriginArgs()
    args.dbfile = '-'
    args.log_level = 3
    args.ip_version = 4
    args.minimum_match = 4
    assert myorigin.my_ip(args) == '4.5.6.7'
    args.minimum_match = 5
    caplog.clear()
    myorigin.my_ip(args)
    assert 'not enough providers for 5 matches' in caplog.text
    args.ip_version = 6
    args.minimum_match = 2
    caplog.clear()
    myorigin.my_ip(args)
    assert 'not enough providers for 2 matches' in caplog.text
    args.minimum_match = 1
    assert myorigin.my_ip(args) == '2604::89'
    myorigin.myorigin.http_get.pytest = 2
    caplog.clear()
    myorigin.my_ip(args)
    assert (
        "https://site43.com/2604::89 → 404 pytest simulated error" in caplog.records[1].message
        or "http://site43.com/2604::89 → 404 pytest simulated error" in caplog.records[1].message
    )
    del myorigin.myorigin.http_get.pytest


def test_my_ip_set2(caplog) -> None:
    myorigin.myorigin.init_db.__dict__.pop('engine', None)  # force Parrot.startup() to re-run
    test_flock_data = '''
        site88.com     site88.com/77                4 s
        site20.com     site20.com/1.2.3.4           6 s
    '''
    myorigin.parrots.Parrot.startup.pytest = textwrap.dedent(test_flock_data)
    myorigin.myorigin.http_get.pytest = 1  # mock http_get()
    args = myorigin.MyoriginArgs()
    args.dbfile = '-'
    args.log_level = 3
    args.ip_version = 4
    args.minimum_match = 5
    caplog.clear()
    myorigin.my_ip(args)
    assert "https://site88.com/77 → Only 2 bytes received" in caplog.records[1].message
    args.ip_version = 6  # IPv6-only, so should not recognize IPv4 address '1.2.3.4'
    caplog.clear()
    myorigin.my_ip(args)
    assert "https://site20.com/1.2.3.4 → No IP address found" in caplog.records[1].message
    del myorigin.myorigin.http_get.pytest


def test_my_ip_set3(caplog) -> None:
    myorigin.myorigin.init_db.__dict__.pop('engine', None)  # force Parrot.startup() to re-run
    test_flock_data = '''
        site48.com     site48.com/127.0.0.1         4 p
        site51.com     site51.com/2b::29            6 s
        site52.com     site52.com/bad::18           6 s
        site53.com     site53.com/2b::29            6 s
        site54.com     site54.com/2b::29            6 s
        site55.com     site55.com/2b::29            6 s
        site56.com     site56.com/2b::29            6 s
        site57.com     site57.com/bad::18           6 s
        site58.com     site58.com/2b::29            6 s
    '''
    myorigin.parrots.Parrot.startup.pytest = textwrap.dedent(test_flock_data)
    myorigin.myorigin.http_get.pytest = 1  # mock http_get()
    args = myorigin.MyoriginArgs()
    args.dbfile = '-'
    args.log_level = 3
    args.ip_version = 4
    args.minimum_match = 5
    caplog.clear()
    myorigin.my_ip(args)
    assert (
        "http://site48.com/127.0.0.1 → Found non-global IP address 127.0.0.1"
        in caplog.records[1].message
    )
    args.log_level = 4
    args.ip_version = 6
    args.majority_ratio = 3  # 2 of 'bad::18', so need 6 to override
    caplog.clear()
    assert myorigin.my_ip(args) == '2b::29'  # rarely a test will succeed without receiving bad::18
    del myorigin.myorigin.http_get.pytest


def test_my_ip_set4(caplog) -> None:
    myorigin.myorigin.init_db.__dict__.pop('engine', None)  # force Parrot.startup() to re-run
    test_flock_data = '''
        site60.com     site60.com/2buo29            4 b
        site61.com     site61.com/127.0.0           4 b
        site62.com     site62.com/3aaf18            4 b
        site63.com     site62.com/bbuo29            4 b
        site64.com     site64.com/r27.0.0           4 b
        site65.com     site65.com/naaf18            4 b
        site66.com     site66.com/lbuo29            4 b
        site67.com     site67.com/                  4 b
        site68.com     site68.com/%20%20%20%20      4 b
        site69.com     site69.com/tttttt            4 b
    '''
    myorigin.parrots.Parrot.startup.pytest = textwrap.dedent(test_flock_data)
    myorigin.myorigin.http_get.pytest = 1  # mock http_get()
    args = myorigin.MyoriginArgs()
    args.dbfile = '-'
    args.log_level = 3
    args.ip_version = 4
    args.max_failures = 9
    caplog.clear()
    myorigin.my_ip(args)
    assert "9 requests failed; giving up" in caplog.records[10].message
    del myorigin.myorigin.http_get.pytest


def test_my_ip_set5(caplog) -> None:
    myorigin.myorigin.init_db.__dict__.pop('engine', None)  # force Parrot.startup() to re-run
    test_flock_data = '''
        site51.com     site51.com/2b::29            6 s
        site52.com     site52.com/bad::18           6 s
        site53.com     site53.com/2b::29            6 s
        site54.com     site54.com/2b::29            6 s
        site55.com     site55.com/2b::29            6 s
        site57.com     site57.com/bad::18           6 s
    '''
    myorigin.parrots.Parrot.startup.pytest = textwrap.dedent(test_flock_data)
    myorigin.myorigin.http_get.pytest = 1  # mock http_get()
    args = myorigin.MyoriginArgs()
    args.dbfile = '-'
    args.minimum_match = 5
    args.log_level = 3
    args.ip_version = 6
    args.majority_ratio = 3  # 2 of 'bad::18', so need 6 to override
    caplog.clear()
    myorigin.my_ip(args)
    assert (
        "multiple IPs received: ['2b::29', 'bad::18']"
        or "multiple IPs received: ['bad::18', '2b::29']"
        in "; ".join([i.message for i in caplog.records])
    )
    del myorigin.myorigin.http_get.pytest


def test_my_ip_set6(caplog, capsys) -> None:
    myorigin.myorigin.init_db.__dict__.pop('engine', None)  # force Parrot.startup() to re-run
    test_flock_data = '''
        site62.com     site62.com/x                 4 p
        site98.com     site98.com/xy                4 s
        site97.com     site97.com/xyx               4 p
        site94.com     site94.com/xyxy              4 s
        site93.com     site93.com/xyxyx             4 p
        site29.com     site29.com/xyxyxy            4 p
        site44.com     site44.com/xyxyxyx           4 p
        site91.com     site91.com/4.5.6.7           4 s
        site80.com     site80.com/4.5.6.7           4 s
        site55.com     site55.com/4.5.6.7           4 s
        site85.com     site85.com/ip_is_4.5.6.7_    4 s
        site43.com     site43.com/2604::89          6 s
    '''
    myorigin.parrots.Parrot.startup.pytest = textwrap.dedent(test_flock_data)
    myorigin.myorigin.http_get.pytest = 1  # mock http_get()
    args = myorigin.MyoriginArgs()
    args.dbfile = '-'
    args.log_level = 3
    args.ip_version = 4
    args.minimum_match = 5
    caplog.clear()
    myorigin.my_ip(args)
    assert 'not enough providers for 5 matches' in caplog.text
    args.ip_version = 6
    args.minimum_match = 2
    caplog.clear()
    myorigin.my_ip(args)
    assert 'not enough providers for 2 matches' in caplog.text
    args.minimum_match = 1
    assert myorigin.my_ip(args) == '2604::89'
    caplog.clear()
    myorigin.parrots.Parrot.show_parrot_db(myorigin.myorigin.init_db.engine)
    captured = capsys.readouterr()
    expected_parrot_db = '''
        ipv6.https://site43.com/2604::89:
            100% success rate (2 of 2)
            290 ms average round trip time
            score: 122 points
        ipv4.https://site91.com/4.5.6.7:
            100% success rate (1 of 1)
            290 ms average round trip time
            score: 122 points
        ipv4.https://site85.com/ip_is_4.5.6.7_:
            100% success rate (1 of 1)
            290 ms average round trip time
            score: 122 points
        ipv4.https://site80.com/4.5.6.7:
            100% success rate (1 of 1)
            290 ms average round trip time
            score: 122 points
        ipv4.https://site55.com/4.5.6.7:
            100% success rate (1 of 1)
            290 ms average round trip time
            score: 122 points
        ipv4.https://site98.com/xy:
            0% success rate (0 of 1)
            most recent error: Only 2 bytes received
            score: 10 points
        ipv4.https://site94.com/xyxy:
            0% success rate (0 of 1)
            most recent error: No IP address found
            score: 10 points
        ipv4.http://site97.com/xyx:
            0% success rate (0 of 1)
            most recent error: No IP address found
            score: 10 points
        ipv4.http://site93.com/xyxyx:
            0% success rate (0 of 1)
            most recent error: No IP address found
            score: 10 points
        ipv4.http://site62.com/x:
            0% success rate (0 of 1)
            most recent error: Only 1 bytes received
            score: 10 points
        ipv4.http://site44.com/xyxyxyx:
            0% success rate (0 of 1)
            most recent error: No IP address found
            score: 10 points
        ipv4.http://site29.com/xyxyxy:
            0% success rate (0 of 1)
            most recent error: No IP address found
            score: 10 points
    '''
    assert captured.out == textwrap.dedent(expected_parrot_db[1:])
    del myorigin.myorigin.http_get.pytest


def test_my_ip_set7(caplog, capsys) -> None:
    while True:  # loop until weighted_random() selects at least one bad IP address
        myorigin.myorigin.init_db.__dict__.pop('engine', None)  # force Parrot.startup() to re-run
        test_flock_data = '''
            site51.com     site51.com/118.4.122.98      4 p
            site52.com     site52.com/170.250.18.80     4 p
            site53.com     site53.com/118.4.122.98      4 p
            site54.com     site54.com/118.4.122.98      4 p
            site55.com     site55.com/118.4.122.98      4 p
            site56.com     site56.com/118.4.122.98      4 p
            site57.com     site57.com/170.250.18.80     4 p
            site58.com     site58.com/118.4.122.98      4 p
        '''
        myorigin.parrots.Parrot.startup.pytest = textwrap.dedent(test_flock_data)
        myorigin.myorigin.http_get.pytest = 1  # mock http_get()
        args = myorigin.MyoriginArgs()
        args.dbfile = '-'
        args.minimum_match = 5
        args.log_level = 4
        args.ip_version = 4
        args.majority_ratio = 3  # 2 of '170.250.18.80', so need 6 to override
        assert (
            myorigin.my_ip(args) == '118.4.122.98'
        )  # rarely a test will succeed without receiving bad::18
        myorigin.parrots.Parrot.show_parrot_db(myorigin.myorigin.init_db.engine)
        captured = capsys.readouterr()
        matches = ""
        expected_parrot_db = '''
            ipv4.http://site51.com/118.4.122.98:
                100% success rate (1 of 1)
                290 ms average round trip time
                score: 122 points
        '''
        t = expected_parrot_db
        matches += "A" if textwrap.dedent(t[1:]) in captured.out else "."
        t = expected_parrot_db.replace('site51', 'site53')
        matches += "B" if textwrap.dedent(t[1:]) in captured.out else "."
        t = expected_parrot_db.replace('site51', 'site54')
        matches += "C" if textwrap.dedent(t[1:]) in captured.out else "."
        t = expected_parrot_db.replace('site51', 'site55')
        matches += "D" if textwrap.dedent(t[1:]) in captured.out else "."
        t = expected_parrot_db.replace('site51', 'site56')
        matches += "E" if textwrap.dedent(t[1:]) in captured.out else "."
        t = expected_parrot_db.replace('site51', 'site58')
        matches += "F" if textwrap.dedent(t[1:]) in captured.out else "."
        expected_parrot_db = '''
            ipv4.http://site52.com/170.250.18.80:
                0% success rate (0 of 1)
                most recent error: Incorrect IP returned: 170.250.18.80
                score: 10 points
        '''
        t = expected_parrot_db
        matches += "G" if textwrap.dedent(t[1:]) in captured.out else "."
        t = expected_parrot_db.replace('site52', 'site57')
        matches += "H" if textwrap.dedent(t[1:]) in captured.out else "."
        dots = matches.count(".")  # a dot represents unmodified site data
        assert dots == 3 or dots == 2 or dots == 0, f"matches is {matches}"
        if "G" in matches or "H" in matches:  # testing sufficient if it included an incorrect IP
            break  # testing sufficient if it included an incorrect IP
    del myorigin.myorigin.http_get.pytest


def test_weighted_random() -> None:
    random.seed('YLLMFHALDNRHISAEIFHIAE')  # predictable 'random' results
    o = {200: 5, 202: 10, 203: 11, 210: 1, 214: 10}
    n = 100  # number of iterations (10000 would make o and r values closer but take much longer)
    h = [k for k in [myorigin.myorigin.weighted_random(o) for i in range(n * sum(o.values()))]]
    r = dict(sorted({i: h.count(i) / n for i in h}.items()))
    assert r == {200: 5.06, 202: 9.85, 203: 10.76, 210: 0.85, 214: 10.48}
    random.seed()  # make unpredictable again so other tests cover a broader range
